
#include "utils.h"
#include <strings.h>
#include <set>

class Part
{
public:
    N weight;
    N length;
    N parts[100];

    Part(N w)
        : weight(w)
        , length(w)
    {
        for(N i=0;i<w;i++)
            parts[i] = 1;
    }

    Part()
        : weight(0)
        , length(0)
    {
    }

    void reset(N w)
    {
        weight = length = w;

        for(N i=0;i<w;i++)
            parts[i] = 1;
    }

    bool next();

    template<N p, N q1, N q2>
    bool next(N pqmax, N* q1rem, N* q2rem);

    bool next(N p, N q1, N q2, N pqmax, N* q1rem, N* q2rem);

    void print();

    void show();

    bool pcore_trivial(N p, bool rem1);

    template<N p, bool rem1>
    bool pcore_trivial();

    template<N d, N p>
    void pquotient(void* q, N* sign);

    void pquotient(N d, N p, void* q, N* sign);

    template<N d, N p, bool rem1>
    void from_pquotient(void* q, N* sign);

    void from_pquotient(N d, N p, bool rem1, void* q, N* sign);

    ZZ prod_hooks();
    
    void prod_hooks(QuickZ& z);
    
    void div_prod_hooks(QuickZ& z);

    N get_multip();
};

Part* nth(void* p, N d, N i)
{
    return (Part*)( (uint8_t*)p + (sizeof(Part)+sizeof(N)*(d-1))*i );
}

template<N d>
Part* nth(void* p, N i)
{
    return (Part*)( (uint8_t*)p + (sizeof(Part)+sizeof(N)*(d-1))*i );
}

/* computes the next partition (in lexicographical order) */
bool Part::next()
{
    for(Z i = length-2; i>=0; i--)
    if(i==0 || (parts[i] < parts[i-1]))
    {
        N rem = parts[length-1] + parts[i] * (length-2-i) - 1;

        parts[i]++;
        for(N j = 0;j<rem;j++)
            parts[i+1+j] = 1;
        length = i+1+rem;
        return true;
    }

    return false;
}

/* computes the next partition, and incrementally updates the data
   to check if q1- and q2-cores are trivial */
template<N p, N q1, N q2>
bool Part::next(N toadd, N* q1rem, N* q2rem)
{
    for(Z i = length-2; i>=0; i--)
    if(i==0 || (parts[i] < parts[i-1]))
    {
        N rem = parts[length-1] + parts[i] * (length-2-i) - 1;

        q1rem[((parts[i]-i)*p + toadd)%q1]--;
        q1rem[((parts[i]+1-i)*p + toadd)%q1]++;
        q2rem[((parts[i]-i)*p + toadd)%q2]--;
        q2rem[((parts[i]+1-i)*p + toadd)%q2]++;
        parts[i]++;

        /* remove all parts */
        for(N j=i+1;j<length;j++)
        {
            q1rem[((parts[j]-j)*p + toadd)%q1]--;
            q1rem[(-j*p + toadd)%q1]++;
            q2rem[((parts[j]-j)*p + toadd)%q2]--;
            q2rem[(-j*p + toadd)%q2]++;
        }

        /* perché p è 1 modulo q1 (fixme) */
        q1rem[(1*p + toadd - p*(i+1))%q1]++;
        q1rem[(toadd - p*(i+rem))%q1]--;
        q2rem[(1*p + toadd - p*(i+1))%q2]++;
        q2rem[(toadd - p*(i+rem))%q2]--;

        for(N j = 0;j<rem;j++)
            parts[i+1+j] = 1;
        length = i+1+rem;
        return true;
    }

    return false;
}

bool Part::next(N p, N q1, N q2, N toadd, N* q1rem, N* q2rem)
{
    for(Z i = length-2; i>=0; i--)
    if(i==0 || (parts[i] < parts[i-1]))
    {
        N rem = parts[length-1] + parts[i] * (length-2-i) - 1;

        q1rem[((parts[i]-i)*p + toadd)%q1]--;
        q1rem[((parts[i]+1-i)*p + toadd)%q1]++;
        q2rem[((parts[i]-i)*p + toadd)%q2]--;
        q2rem[((parts[i]+1-i)*p + toadd)%q2]++;
        parts[i]++;

        /* remove all parts */
        for(N j=i+1;j<length;j++)
        {
            q1rem[((parts[j]-j)*p + toadd)%q1]--;
            q1rem[(-j*p + toadd)%q1]++;
            q2rem[((parts[j]-j)*p + toadd)%q2]--;
            q2rem[(-j*p + toadd)%q2]++;
        }

        /* perché p è 1 modulo q1 (fixme) */
        q1rem[(1*p + toadd - p*(i+1))%q1]++;
        q1rem[(toadd - p*(i+rem))%q1]--;
        q2rem[(1*p + toadd - p*(i+1))%q2]++;
        q2rem[(toadd - p*(i+rem))%q2]--;

        for(N j = 0;j<rem;j++)
            parts[i+1+j] = 1;
        length = i+1+rem;
        return true;
    }

    return false;
}

/* initializes the data to quickly check if a partitions is a q-core */
void init_prem(N d, N p, N q, void* quot, N* p1rem)
{
    for(N i=0;i<q;i++)
        p1rem[i] = 0;

    for(N i=0;i<p;i++)
    {
        Z adj = (d % p) ? (i==0?1:(i==p-1?-1:0)) : 0;
        for(Z j=0;j<(Z)(d/p+1)+adj;j++)
        {
            N part = j < (Z)nth(quot, d, i)->length ? nth(quot, d, i)->parts[j] : 0;
            
            p1rem[((part + d/p+1 + adj - 1 - j)*p + i)%q]++;        
        }
    }

    /* pre-apply adjustment to quickly check if q-core is trivialOB */
    N d_mod_p = ((d/p+1)*p) % q;
    for(N i=0; i < d_mod_p; i++)
        p1rem[i]--;
    if(d % q)
    {    
        p1rem[d_mod_p == 0 ? q-1 : d_mod_p-1]++;
        p1rem[d_mod_p]--;
    }
}

/* verify if it is a q-core */
template<N q>
bool check_pcore(N* p1rem)
{
    for(N i=1;i<q;i++)
    if(p1rem[0] != p1rem[i])
        return false;
    return true;
}

bool check_pcore(N q, N* p1rem)
{
    for(N i=1;i<q;i++)
    if(p1rem[0] != p1rem[i])
        return false;
    return true;
}

template<N d, N p, N q1, N q2>
bool next_pquotient(void* quot, N* q1rem, N* q2rem)
{
    const bool rem1 = d % p;
    N toadd_pm1 = (d/p+1 - 1 - rem1)*p + (p-1); /* ( (delta_1 - qquot_1)*p + r ) % q */
    Part *q = nth<d>(quot, p-1);
    if(q->next<p,q1,q2>(toadd_pm1, q1rem, q2rem))
        return true;

    q1rem[(nth<d>(quot, p-1)->weight*p + toadd_pm1 - 0)%q1]--;
    q1rem[(toadd_pm1 - 0)%q1]++;
    q2rem[(nth<d>(quot, p-1)->weight*p + toadd_pm1 - 0)%q2]--;
    q2rem[(toadd_pm1 - 0)%q2]++;


    N remaining = nth<d>(quot, p-1)->weight;
    Z i = p-2;
    while(i>=0 && !nth<d>(quot, i)->weight)
        i--;

    if(i==-1)
        return false;

    /* either step on, or remove an element and reset */
    N toadd_i = (d/p+1 - 1 + (rem1 && i==0)) * p + i;
    Part *qi = nth<d>(quot, i);
    if(!qi->next<p,q1,q2>(toadd_i, q1rem, q2rem))
    {
        q1rem[(nth<d>(quot, i)->weight*p + toadd_i - 0)%q1]--;
        q1rem[(toadd_i - 0)%q1]++;
        q2rem[(nth<d>(quot, i)->weight*p + toadd_i - 0)%q2]--;
        q2rem[(toadd_i - 0)%q2]++;

        remaining++;
        nth<d>(quot, i)->reset(nth<d>(quot, i)->weight - 1);

        q1rem[(1*p + toadd_i - 0)%q1]++;
        q1rem[(toadd_i - p*(nth<d>(quot, i)->length-1))%q1]--;
        q2rem[(1*p + toadd_i - 0)%q2]++;
        q2rem[(toadd_i - p*(nth<d>(quot, i)->length-1))%q2]--;
    }

    /* now reset the (i+1)-partition, with remaining elements */
    nth<d>(quot, p-1)->reset(0);
    nth<d>(quot, i+1)->reset(remaining);

    N toadd_ip1 = (d/p+1 - 1 - (rem1 && (i+1==p-1)))*p + (i + 1);
    q1rem[(1*p + toadd_ip1 - 0)%q1]++;
    q1rem[(toadd_ip1 - p*(remaining-1))%q1]--;
    q2rem[(1*p + toadd_ip1 - 0)%q2]++;
    q2rem[(toadd_ip1 - p*(remaining-1))%q2]--;
    return true;
}

bool next_pquotient(N d, N p, N q1, N q2, void* quot, N* q1rem, N* q2rem)
{
    const bool rem1 = d % p;
    N toadd_pm1 = (d/p+1 - 1 - rem1)*p + (p-1); /* ( (delta_1 - qquot_1)*p + r ) % q */
    if(nth(quot, d, p-1)->next(p,q1,q2, toadd_pm1, q1rem, q2rem))
        return true;

    q1rem[(nth(quot, d, p-1)->weight*p + toadd_pm1 - 0)%q1]--;
    q1rem[(toadd_pm1 - 0)%q1]++;
    q2rem[(nth(quot, d, p-1)->weight*p + toadd_pm1 - 0)%q2]--;
    q2rem[(toadd_pm1 - 0)%q2]++;


    N remaining = nth(quot, d, p-1)->weight;
    Z i = p-2;
    while(i>=0 && !nth(quot, d, i)->weight)
        i--;

    if(i==-1)
        return false;

    /* either step on, or remove an element and reset */
    N toadd_i = (d/p+1 - 1 + (rem1 && i==0)) * p + i;
    if(!nth(quot, d, i)->next(p,q1,q2, toadd_i, q1rem, q2rem))
    {
        q1rem[(nth(quot, d, i)->weight*p + toadd_i - 0)%q1]--;
        q1rem[(toadd_i - 0)%q1]++;
        q2rem[(nth(quot, d, i)->weight*p + toadd_i - 0)%q2]--;
        q2rem[(toadd_i - 0)%q2]++;

        remaining++;
        nth(quot, d, i)->reset(nth(quot, d, i)->weight - 1);

        q1rem[(1*p + toadd_i - 0)%q1]++;
        q1rem[(toadd_i - p*(nth(quot, d, i)->length-1))%q1]--;
        q2rem[(1*p + toadd_i - 0)%q2]++;
        q2rem[(toadd_i - p*(nth(quot, d, i)->length-1))%q2]--;
    }

    /* now reset the (i+1)-partition, with remaining elements */
    nth(quot, d, p-1)->reset(0);
    nth(quot, d, i+1)->reset(remaining);

    N toadd_ip1 = (d/p+1 - 1 - (rem1 && (i+1==(Z)(p-1))))*p + (i + 1);
    q1rem[(1*p + toadd_ip1 - 0)%q1]++;
    q1rem[(toadd_ip1 - p*(remaining-1))%q1]--;
    q2rem[(1*p + toadd_ip1 - 0)%q2]++;
    q2rem[(toadd_ip1 - p*(remaining-1))%q2]--;
    return true;
}

void Part::print()
{
    printf("(");
    for(N i = 0;i<length-1;i++)
        printf("%u,", parts[i]);
    printf("%u)\n", parts[length-1]);
}

void Part::show()
{
    for(N i = 0;i<length;i++)
    {
        printf("     ");
        for(N j = 0; j<parts[i]; j++)
            printf("X");
        printf("\n");
    }
}

template<N p, bool rem1>
bool Part::pcore_trivial()
{
    N mod_p[p];
    for(N i = 0; i<p; i++)
        mod_p[i] = 0;

    for(N i = 0; i<length; i++)
        mod_p[(parts[i] + length-1-i) % p]++;
    
    /* if the partition is has a trivial p-core, the classes of l_i+d_i must cover
       all classes modulo p an equal number of time, except of the classes from 0
       to (length mod p) excluded, which exceed the number of the other classes by 1.
       If the p-core is (1), the last one is moved forward by one class */
    N lmod_p = length % p;
    for(N i = 0; i < lmod_p; i++)
        mod_p[i]--;
    if(rem1)
    {
        mod_p[lmod_p == 0 ? p-1 : lmod_p-1]++;
        mod_p[lmod_p]--;
    }
    
    /* now check if there is an equal number in each class */
    for(N i=1;i<p;i++)
    if(mod_p[0] != mod_p[i])
        return false;

    return true;
}

template<N d, N p>
void Part::pquotient(void* quot, N* sign)
{
    for(N i = 0; i<p; i++)
        nth<d>(quot, i)->weight = nth<d>(quot, i)->length = 0;

    N* pdone = (N*)alloca(p * sizeof(N));
    for(N i = 0; i<p; i++)
        pdone[i] = 0;

    for(N i = 0; i<length; i++)
    {
        N c = parts[i] + length-1-i;
        N c_mod_p = c % p;
        N part = (c - c_mod_p) / p; /* part of p-quotient + delta_pq */

        pdone[c_mod_p]++;

        Part *q = nth<d>(quot, c_mod_p);
        q->parts[q->length] = part;
        q->length++;
        q->weight += part;
    }

    /* subtract delta from each part */
    for(N i = 0; i<p; i++)
    {
        Part *q = nth(quot, d, i);

        for(N j = 0; j<q->length; j++)
        {
            N di = q->length-1-j;
            q->parts[j] -= di;
            q->weight -= di;
        }

        while(q->length && q->parts[q->length-1] == 0)
            q->length--;
    }

    N* perm = (N*)alloca(length * sizeof(N));

    for(N i = 0; i<length; i++)
    {
        N c = parts[i] + length-1-i;
        N c_mod_p = c % p;
        N to = (pdone[c_mod_p]-1)*p + c_mod_p; /* final value */
        pdone[c_mod_p]--;

        perm[length-1-i] = to >= length-1 ? length-1 : to;
    }

    /* now we just have to compute the sign of "perm" */
    *sign = perm_sign(length, perm);
}

void Part::pquotient(N d, N p, void* quot, N* sign)
{
    for(N i = 0; i<p; i++)
        nth(quot, d, i)->weight = nth(quot, d, i)->length = 0;

    N* pdone = (N*)alloca(p * sizeof(N));
    for(N i = 0; i<p; i++)
        pdone[i] = 0;

    for(N i = 0; i<length; i++)
    {
        N c = parts[i] + length-1-i;
        N c_mod_p = c % p;
        N part = (c - c_mod_p) / p; /* part of p-quotient + delta_pq */

        pdone[c_mod_p]++;

        Part *q = nth(quot, d, c_mod_p);
        q->parts[q->length] = part;
        q->length++;
        q->weight += part;
    }

    /* subtract delta from each part */
    for(N i = 0; i<p; i++)
    {
        Part *q = nth(quot, d, i);

        for(N j = 0; j<q->length; j++)
        {
            N di = q->length-1-j;
            q->parts[j] -= di;
            q->weight -= di;
        }

        while(q->length && q->parts[q->length-1] == 0)
            q->length--;
    }

    N* perm = (N*)alloca(length * sizeof(N));

    for(N i = 0; i<length; i++)
    {
        N c = parts[i] + length-1-i;
        N c_mod_p = c % p;
        N to = (pdone[c_mod_p]-1)*p + c_mod_p; /* final value */
        pdone[c_mod_p]--;

        perm[length-1-i] = to >= length-1 ? length-1 : to;
    }

    /* now we just have to compute the sign of "perm" */
    *sign = perm_sign(length, perm);
}

template<N d, N p, bool rem1>
void Part::from_pquotient(void* quot, N* sign)
{
    N used[p];
    for(N i = 0; i<p; i++)
        used[i] = 0;

    N maxqlen = 0;
    for(N i=0;i<p;i++)
        maxqlen = std::max(maxqlen, nth<d>(quot, i)->length);
    maxqlen++;
    N delta_len = maxqlen * p;
    
    N perm[maxqlen * p];
    Z perm_idx = maxqlen * p - 1;

    weight = length = 0;
    Z prev = -1000000;

    while(true)
    {
        N bidx = 0;
        Z psub = 0;
        Z biggest = -1000001;
        for(Z i = p-1; i >= 0; i--)
        {
            Part *q = nth<d>(quot, i);
            Z qpart = used[i]>=q->length ? 0 : q->parts[used[i]];
            Z pdelta = - used[i] + (rem1 ? (i==0 ? 1 : ((i==p-1) ? -1 : 0)) : 0);
            Z qdpart = qpart + pdelta; /* pdelta + qcore */
            if(qdpart > biggest)
            {
                bidx = i;
                biggest = qdpart;
                psub = pdelta;
                if(biggest == prev)
                    break;
            }
        }

        prev = biggest;

        /* where this part should stay */
        perm[perm_idx--] = std::min(bidx + (psub + maxqlen - 1)*p, delta_len -1);
        used[bidx]++;

        /* (part from the p-core) + delta */
        N part = (bidx + (biggest + maxqlen - 1)*p) - (delta_len-1-length); 

        if(!part)
            break;
        parts[length++] = part;
        weight += part;
    }

    while(perm_idx >= 0)
    {
        perm[perm_idx] = perm_idx;
        perm_idx--;
    }

    /* now we just have to compute the sign of "perm" */
    *sign = perm_sign(maxqlen * p, perm);
}

void Part::from_pquotient(N d, N p, bool rem1, void* quot, N* sign)
{
    N used[p];
    for(N i = 0; i<p; i++)
        used[i] = 0;

    N maxqlen = 0;
    for(N i=0;i<p;i++)
        maxqlen = std::max(maxqlen, nth(quot, d, i)->length);
    maxqlen++;
    N delta_len = maxqlen * p;
    
    N perm[maxqlen * p];
    Z perm_idx = maxqlen * p - 1;

    weight = length = 0;
    Z prev = -1000000;

    while(true)
    {
        N bidx = 0;
        Z psub = 0;
        Z biggest = -1000001;
        for(Z i = p-1; i >= 0; i--)
        {
            Part *q = nth(quot, d, i);
            Z qpart = used[i]>=q->length ? 0 : q->parts[used[i]];
            Z pdelta = - used[i] + (rem1 ? (i==0 ? 1 : ((i==(Z)(p-1)) ? -1 : 0)) : 0);
            Z qdpart = qpart + pdelta; /* pdelta + qcore */
            if(qdpart > biggest)
            {
                bidx = i;
                biggest = qdpart;
                psub = pdelta;
                if(biggest == prev)
                    break;
            }
        }

        prev = biggest;

        /* where this part should stay */
        perm[perm_idx--] = std::min(bidx + (psub + maxqlen - 1)*p, delta_len -1);
        used[bidx]++;

        /* (part from the p-core) + delta */
        N part = (bidx + (biggest + maxqlen - 1)*p) - (delta_len-1-length); 

        if(!part)
            break;
        parts[length++] = part;
        weight += part;
    }

    while(perm_idx >= 0)
    {
        perm[perm_idx] = perm_idx;
        perm_idx--;
    }

    /* now we just have to compute the sign of "perm" */
    *sign = perm_sign(maxqlen * p, perm);
}

void Part::prod_hooks(QuickZ& z)
{
    for(N i=0;i<length;i++)
    {
        N part = parts[i];
        N next_part = i==length-1 ? 0 : parts[i+1];
        
        if(part == next_part)
            continue;

        for(N x = part, y = 0; y < i; y++)
        {
            N sq_down = i - y;
            N sq_right = parts[y] - x;
            z *= (sq_right + sq_down + 1);
        }

        for(N x = part-1; x>next_part; x--)
        for(N y = 0; y <= i; y++)
        {
            N sq_down = i - y;
            N sq_right = parts[y] - x;
            z *= (sq_right + sq_down + 1);
        }
    }
}

void Part::div_prod_hooks(QuickZ& z)
{
    for(N i=0;i<length;i++)
    {
        N part = parts[i];
        N next_part = i==length-1 ? 0 : parts[i+1];
        
        if(part == next_part)
            continue;

        for(N x = part, y = 0; y < i; y++)
        {
            N sq_down = i - y;
            N sq_right = parts[y] - x;
            z /= (sq_right + sq_down + 1);
        }

        for(N x = part-1; x>next_part; x--)
        for(N y = 0; y <= i; y++)
        {
            N sq_down = i - y;
            N sq_right = parts[y] - x;
            z /= (sq_right + sq_down + 1);
        }
    }
}

/* Returns a multiplier, which is 1 for simmetric partitions, and for
   each pair of conjugate partitions m, m', is 2 for the lexicographically
   bigger, and 0 for the other one. To speed up the computation. */
N Part::get_multip()
{
    if(length > parts[0])
        return 0;
    if(parts[0] > length)
        return 2;

    N l = length;
    for(N pl = 1; pl < length; pl++)
    {
        while(l && parts[l-1]<=pl)
            l--;
        if(l > parts[pl])
            return 0;
        if(parts[pl] > l)
            return 2;
        if(l < pl)
            return 1;
    }

    return 1;
}

void conjugate(const Part& from, Part& to)
{
    to.length = from.parts[0];
    to.parts[0] = from.length;

    N l = from.length;
    for(N pl = 1; pl < to.length; pl++)
    {
        while(l && from.parts[l-1]<=pl)
            l--;
        to.parts[pl] = l;
    }
}

template<N d, N p>
void div_hook_lengths(void* pquot, N sign, QuickZ& z)
{
    for(N i=0;i<p;i++)
        nth<d>(pquot, i)->div_prod_hooks(z);
        
    if(sign & 1)
        z.flip_sign();
}

void div_hook_lengths(N d, N p, void* pquot, N sign, QuickZ& z)
{
    for(N i=0;i<p;i++)
        nth(pquot, d, i)->div_prod_hooks(z);
        
    if(sign & 1)
        z.flip_sign();
}

/* computes the nth coefficients of P(x)^k, where P(x)
   is the partitions generating function*/
uint64_t num_kpartitions(int k, int n)
{
    uint64_t p[n+1];

    for(int i=0;i<=n;i++)
        p[i] = 1;

    /* sets p[0],p[1],... to the coefficients of P(x), up to x^n */
    for(int i=2;i<=n;i++)
    {        
        uint64_t p2[n+1];
        for(int j=0;j<=n;j++)
            p2[j] = p[j];

        for(int d = i; d <= n; d += i)
        for(int j=d;j<=n;j++)
            p2[j] += p[j-d];

        for(int j=0;j<=n;j++)
            p[j] = p2[j];
    }

    /* dummy exponentiation to P(x)^k */
    uint64_t r[n+1];
    for(int j=0;j<=n;j++)
        r[j] = p[j];
    for(int e=1;e<k;e++)
    {
        uint64_t p2[n+1];
        for(int i=0; i<=n;i++)
        {
            p2[i] = 0;
            for(int j=0;j<=i;j++)
                p2[i] += p[j] * r[i-j];
        }
        for(int j=0;j<=n;j++)
            r[j] = p2[j];
    }

    return r[n];
}


typedef std::set<N64> DegSet;

void compute_degrees(N d)
{
    Part& p = *(Part*)alloca(sizeof(Part) + sizeof(N)*(d-1));
    p.reset(d);
    ZZ fac = factorial(d), deg;
    QuickZ hooks;
    QuickZ::set_d(d);
    DegSet degs;

    do
    {
        if(!p.get_multip())
            continue;

        hooks.reset(1);
        p.prod_hooks(hooks);
        hooks.normalize();

        deg = fac / hooks.zz;
        degs.insert( deg.get_ui() );
    }
    while(p.next());

    printf("\n-------------------------------------------------------------\n");
    printf("ch(S_%d) = { ", d);
    N num = 1;
    N d2, d3, d9 = 0;
    for(DegSet::iterator it = degs.begin(); it != degs.end(); ++it)
    {
        if(num==2)
            d2 = *it;
        else if(num==3)
            d3 = *it;
        else if(num==9)
            d9 = *it;
        if(num!=1 && num%8==1)
            printf("\n    ");
        num++;
        printf("%s %lu", (it!=degs.begin()?",":""), *it);
    }
    printf(" }\n");
    printf("-------------------------------------------------------------\n");
    if(d9 != 0 && d>=20)
    {
        bool vera = d2*d3 < d9;
        printf("Selecongettura: %s\n", vera? "vera!" : "falsa :-(");
        if(!vera)
            exit(0);
    }
}

void go_on()
{
    for(N i = 2; i<50; i++)
        compute_degrees(i);
}

const N s_d = Settings::d;
const N s_p1 = Sort3<Settings::p1, Settings::p2, Settings::p3>::s_biggest;
const N s_p2 = Sort3<Settings::p1, Settings::p2, Settings::p3>::s_middle;
const N s_p3 = Sort3<Settings::p1, Settings::p2, Settings::p3>::s_smallest;

ZNat QuickZ::s_test_overflow = QuickZ::s_test_overflow_default;

N d = s_d;
N p1 = s_p1;
N p2 = s_p2;
N p3 = s_p3;


int main(int argc, char *argv[])
{
    bool quick_exit = false;
    if(argc >= 2 && !strcasecmp(argv[1], "dummy"))
    {
        argc--;
        argv++;
        quick_exit = true;
    }

    if(argc >= 2 && !strcasecmp(argv[1], "-b"))
    {
        go_on();
        return 0;
    }
    
    if((argc >= 2 && (!strcasecmp(argv[1], "-h") ||
                     !strcasecmp(argv[1], "-help") ||
                     !strcasecmp(argv[1], "--h") ||
                      !strcasecmp(argv[1], "--help")))
       || (argc != 1 && argc != 5) )
    {
        printf("Help: run with the integers d p1 p2 p3 as arguments,\n"
               "  or with no argument to use the defaults in settings.h\n");
        return 0;
    }

    bool custom_parameters = false;

    if(argc == 5)
    {
        bool valid = true;
        char *ptr;
        N pi[3];
        d  = strtol(argv[1], &ptr, 0); valid = valid && (*ptr == '\0');
        pi[0] = strtol(argv[2], &ptr, 0); valid = valid && (*ptr == '\0');
        pi[1] = strtol(argv[3], &ptr, 0); valid = valid && (*ptr == '\0');
        pi[2] = strtol(argv[4], &ptr, 0); valid = valid && (*ptr == '\0');

        std::sort(pi, pi+3);
        p1 = pi[2];
        p2 = pi[1];
        p3 = pi[0];

        if(!valid || d <= 0 || p1 <= 0 || p2 <= 0 || p3 <= 0)
        {
            printf("Error: The parameters must be positive integers!!\n");
            return 0;
        }

        if(s_d != d || s_p1 != p1 || s_p2 != p2 || s_p3 != p3)
            custom_parameters = true;
    }

    if(d%p1 != 0 && d%p1 != 1)
    {
        printf("The degree (%u) should be  = 0 or 1 (mod %u)!!!\n", d, p1);
        return 0;
    }
    if(d%p2 != 0 && d%p2 != 1)
    {
        printf("The degree (%u) should be  = 0 or 1 (mod %u)!!!\n", d, p2);
        return 0;
    }
    if(d%p3 != 0 && d%p3 != 1)
    {
        printf("The degree (%u) should be  = 0 or 1 (mod %u)!!!\n", d, p3);
        return 0;
    }
    if( ((p1+1)*(d/p1) + (p2+1)*(d/p2) + (p3+1)*(d/p3) )% 2 != 0)
    {
        printf("No solution, there is an odd number of permutations\n"
               "  in S_d\\A_d, their product cannot be 1!!\n");
        printf("RESULT = 0\n");
        return 0;
    }

    /* partition, and counters */
    unsigned long long n = 0, t = 0, prevn = 0;
    Part& p = *(Part*)alloca(sizeof(Part) + sizeof(N)*(d-1));

    /* initialize the p1-quotients, which will be enumerated */
    void* p1_quot = (Part*)alloca((sizeof(Part) + sizeof(N)*(d-1)) * p1);
    void* p2_quot = (Part*)alloca((sizeof(Part) + sizeof(N)*(d-1)) * p2);
    void* p3_quot = (Part*)alloca((sizeof(Part) + sizeof(N)*(d-1)) * p3);
    nth(p1_quot, d, 0)->reset(d/p1);
    for(N i=1;i<p1;i++)
        nth(p1_quot, d, i)->reset(0);

    /* initialize the array that will be used to quickly
       check if the partition given by the p1-quotient has
       a trivial p2- or p3-core */
    N p2rem[p2];
    N p3rem[p3];
    init_prem(d, p1, p2, p1_quot, p2rem);
    init_prem(d, p1, p3, p1_quot, p3rem);

    ZZ sum = 0;
    QuickZ phooks, pchars;
    ZZ base_p1 = factorial(d/p1);
    ZZ base_p2 = factorial(d/p2);
    ZZ base_p3 = factorial(d/p3);
    uint64_t start_time = current_time();


    /* test code (slow) */
#if 0
    /* this is the number of p-quotients, just to print the status */
    unsigned long long numquot = num_kpartitions(1, d);
    p.reset(d);

    do
    {
        n++;

        if(!p.pcore_trivial<p1, d%p1>())
            continue;

        if(!p.pcore_trivial<p2, d%p2>())
            continue;

        if(!p.pcore_trivial<p3, d%p3>())
            continue;

        t++;

        N sign_p1, sign_p2, sign_p3;
        p.pquotient(p1, p1_quot, &sign_p1);
        p.pquotient(p2, p2_quot, &sign_p2);
        p.pquotient(p3, p3_quot, &sign_p3);


        /* compute che characters */
        phooks.reset();
        p.prod_hooks(phooks);
        //phooks *= multip;
        phooks.normalize();

        pchars.reset(base_p1);
        div_hook_lengths(d, p1, p1_quot, sign_p1, pchars);

        pchars.zz *= base_p2;
        div_hook_lengths(d, p2, p2_quot, sign_p2, pchars);

        pchars.zz *= base_p3;
        div_hook_lengths(d, p3, p3_quot, sign_p3, pchars);
        pchars.div_normalize();

        /* sum the product of characters */
        sum += phooks.zz * pchars.zz;

        if(n > prevn + 2000000)
        {
            prevn = n;

            printf("\n\n");
            p.show();
            p.print();
            printf("partial sum = %s\n", sum.get_str().c_str());
            printf("%llu/%llu, %llu computed\n", n, numquot, t);

            uint64_t elap = current_time() - start_time;
            uint64_t eta = elap * (numquot-n) / n;
            char buf[256], buf2[256];
            printf("\n~%02llu%% done     elapsed time: %s     eta: %s\n",
                   100*n/numquot,
                   format_time(buf, elap),
                   format_time(buf2, eta) );

            if(argc > 1 && n>8000000)
                return 0;
        }
    }
    while(p.next());
#else


    if(!custom_parameters)
    {
        /* this is the number of p-quotients, just to print the status */
        unsigned long long numquot = num_kpartitions(p1, d/p1);

        do
        {
            n++;

#if 1
            /* verify if the p2-core is (almost) trivial */
            if(!check_pcore<s_p2>(p2rem))
                continue;

            /* verify if the p3-core is (almost) trivial */
            if(!check_pcore<s_p3>(p3rem))
                continue;
#endif

            /* get the partitions from its p1-core */
            N sign_p1;
            p.from_pquotient<s_d, s_p1, s_d % s_p1>(p1_quot, &sign_p1);

#if 0
            /* a sanity check, uncomment if you suspect the presence of a bug,
               and comment the exit conditions above */
            N p2remW[p2];
            N p3remW[p3];
            init_prem<p1, p2>(p1_quot, p2remW);
            init_prem<p1, p3>(p1_quot, p3remW);
            for(int i=0;i<p2;i++)
                if(p2rem[i] != p2remW[i])
                    TRAP;
            for(int i=0;i<p3;i++)
                if(p3rem[i] != p3remW[i])
                    TRAP;
            if(check_pcore<p2>(p2rem) != p.pcore_trivial<p2, d%p2>())
            {
                int a = check_pcore<p1, p2>(p2rem);
                int b = p.pcore_trivial<p2, d%p2>();
                printf("%d %d\n", a, b);
                TRAP;
                check_pcore<p1, p2>(p2rem);
                p.pcore_trivial<p2, d%p2>();
            }
            if(check_pcore<p3>(p3rem) != p.pcore_trivial<p3, d%p3>())
                TRAP;
            if(!check_pcore<p2>(p2rem))
                continue;
            if(!check_pcore<p3>(p3rem))
                continue;
#endif

            /* compute only once for each pair of conjugate partitions */
            Z multip = p.get_multip();
            if(!multip)
                continue;

            t++;

            /* compute the p2- and p3-quotients */
            N sign_p2, sign_p3;
            p.pquotient<s_d, s_p2>(p2_quot, &sign_p2);
            p.pquotient<s_d, s_p3>(p3_quot, &sign_p3);

            /* compute the product of hook-lengths values */
            phooks.reset(multip);
            p.prod_hooks(phooks);
            phooks.normalize();

            /* compute the product of the character values,
               evaluated on our  conjugacy classes */
            pchars.reset(base_p1);
            div_hook_lengths<s_d, s_p1>(p1_quot, sign_p1, pchars);

            pchars.zz *= base_p2;
            div_hook_lengths<s_d, s_p2>(p2_quot, sign_p2, pchars);

            pchars.zz *= base_p3;
            div_hook_lengths<s_d, s_p3>(p3_quot, sign_p3, pchars);
            pchars.div_normalize();

            /* add the product of characters and hook lengths */
            sum += phooks.zz * pchars.zz;

            if(n > prevn + 2000000)
            {
                prevn = n;

                printf("\n\n");
                p.show();
                p.print();
                printf("partial sum = %s\n", sum.get_str().c_str());
                printf("%llu/%llu, %llu computed\n", n, numquot, t);

                uint64_t elap = current_time() - start_time;
                uint64_t eta = elap * (numquot-n) / n;
                char buf[256], buf2[256];
                printf("\n~%02llu%% done     elapsed time: %s     eta: %s\n",
                       100*n/numquot,
                       format_time(buf, elap),
                       format_time(buf2, eta) );

                if(quick_exit && n>8000000)
                    return 0;
            }
        }
        while(next_pquotient<s_d, s_p1, s_p2, s_p3>(p1_quot, p2rem, p3rem));
    }
    else
    {
        /* this is the number of p-quotients, just to print the status */
        unsigned long long numquot = num_kpartitions(p1, d/p1);
        QuickZ::set_d(d);

        do
        {
            n++;

            /* verify if the p2-core is (almost) trivial */
            if(!check_pcore(p2, p2rem))
                continue;

            /* verify if the p3-core is (almost) trivial */
            if(!check_pcore(p3, p3rem))
                continue;

            /* get the partitions from its p1-core */
            N sign_p1;
            p.from_pquotient(d, p1, d%p1, p1_quot, &sign_p1);

            /* compute only once for each pair of conjugate partitions */
            Z multip = p.get_multip();
            if(!multip)
                continue;

            t++;

            /* compute the p2- and p3-quotients */
            N sign_p2, sign_p3;
            p.pquotient(d, p2, p2_quot, &sign_p2);
            p.pquotient(d, p3, p3_quot, &sign_p3);

            /* compute the product of hook-lengths values */
            phooks.reset(multip);
            p.prod_hooks(phooks);
            phooks.normalize();

            /* compute the product of the character values,
               evaluated on our  conjugacy classes */
            pchars.reset(base_p1);
            div_hook_lengths(d, p1, p1_quot, sign_p1, pchars);

            pchars.zz *= base_p2;
            div_hook_lengths(d, p2, p2_quot, sign_p2, pchars);

            pchars.zz *= base_p3;
            div_hook_lengths(d, p3, p3_quot, sign_p3, pchars);
            pchars.div_normalize();

            /* add the product of characters and hook lengths */
            sum += phooks.zz * pchars.zz;

            if(n > prevn + 2000000)
            {
                prevn = n;

                printf("\n\n");
                p.show();
                p.print();
                printf("partial sum = %s\n", sum.get_str().c_str());
                printf("%llu/%llu, %llu computed\n", n, numquot, t);

                uint64_t elap = current_time() - start_time;
                uint64_t eta = elap * (numquot-n) / n;
                char buf[256], buf2[256];
                printf("\n~%02llu%% done     elapsed time: %s     eta: %s\n",
                       100*n/numquot,
                       format_time(buf, elap),
                       format_time(buf2, eta) );

                if(quick_exit && n>8000000)
                    return 0;
            }
        }
        while(next_pquotient(d, p1, p2, p3, p1_quot, p2rem, p3rem));
    }
#endif

    /* z_lambda^-1 is equal to |C_lambda|/d! */
    ZZ z_p1 = exp(p1, d/p1) * factorial(d/p1);
    ZZ z_p2 = exp(p2, d/p2) * factorial(d/p2);
    ZZ z_p3 = exp(p3, d/p3) * factorial(d/p3);
    ZZ z = z_p1 * z_p2 * z_p3;
    ZZ zdivgcd = z / gcd(z_p3, gcd(z_p2, z_p1));
    ZZ num_triples = sum * factorial(d) / z;

    /* print results */
    printf("\n------------------------------------------------------------------------\n");
    printf("Generated %llu %d-quotients, %llu characters evaluated.\n", n, p1, t);
    printf("\nSUM    = %s\n", sum.get_str().c_str());
    printf("(this is the sum of products of characters and hook-lengths, not normalized)\n");

    printf("\nRESULT = %s\n", num_triples.get_str().c_str());
    printf("(this is the number of solutions to the equation s1*s2*s3 = 1 in S_%u)\n", d);
    
    printf("\nZ/GCD  = %s\n", zdivgcd.get_str().c_str());
    printf("\nS %% (Z/GCD) = %s\n", ZZ(sum % zdivgcd).get_str().c_str());
    printf("(since each conjugation-orbit of triples has cardinality multiple of d!/gcd(z_{\\mu_i}),\n"
           " this remainder should be 0, or there is a bug in the program!)\n");
    printf("\n------------------------------------------------------------------------\n");

    return 0;
}
